@@ -34,6 +34,7 @@
     `(progn
        (declaim (inline ,function))
        (defun ,function ,args
+         (declare (sb-c:flushable sb-c:%alien-funcall))
          (truly-the ;; avoid checking the result
           ,(type-specifier (fun-type-returns (info :function :type function)))
           (alien-funcall
@@ -76,6 +77,7 @@
 #-x86 (def-math-rtn "atan" 1)
 #-x86 (def-math-rtn "atan2" 2)
 
+;;; See src/runtime/wrap.c for the definitions of the "sb_"-prefixed things.
 (def-math-rtn "acos" 1 #+win32 t)
 (def-math-rtn "asin" 1 #+win32 t)
 (def-math-rtn "cosh" 1 #+win32 t)
@@ -119,6 +121,12 @@
          (if (evenp power)
              1
              base))
+        ((eql base 0)
+         (cond ((= power 0) 1)
+               ((> power 0) 0)
+               (t (error 'division-by-zero
+                         :operands (list 0 power)
+                         :operation 'expt))))
         ((ratiop base)
          (let ((den (denominator base))
                (num (numerator base)))
@@ -253,7 +261,7 @@
                            (when (< x-hi 0)
                              (cond ((and (= x-ihi #x3ff00000) (zerop yisint))
                                     ;; (-1)**non-int
-                                    (let ((y*pi (* y sb-xc:pi)))
+                                    (let ((y*pi (* y pi)))
                                       (declare (double-float y*pi))
                                       (return-from real-expt
                                         (complex
@@ -276,7 +284,7 @@
                                (2 ; even
                                 (coerce pow rtype))
                                (t ; non-integer
-                                (let ((y*pi (* y sb-xc:pi)))
+                                (let ((y*pi (* y pi)))
                                   (declare (double-float y*pi))
                                   (complex
                                    (coerce (* pow (%cos y*pi))
@@ -334,6 +342,7 @@
 
 ;;; FIXME: Maybe rename this so that it's clearer that it only works
 ;;; on integers?
+(declaim (inline log2))
 (defun log2 (x)
   (declare (type integer x))
   ;; CMUCL comment:
@@ -343,15 +352,19 @@
   ;;   appropriately, take the log of it and add it to n.
   ;;
   ;; Motivated by an attempt to get LOG to work better on bignums.
-  (let ((n (integer-length x)))
-    (if (< n sb-vm:double-float-digits)
-        (log (coerce x 'double-float) $2.0d0)
-        (let ((f (ldb (byte sb-vm:double-float-digits
-                            (- n sb-vm:double-float-digits))
-                      x)))
-          (+ n (log (scale-float (coerce f 'double-float)
-                                 (- sb-vm:double-float-digits))
-                    $2.0d0))))))
+  (cond ((typep x 'sb-vm:signed-word)
+         (log (coerce x 'double-float) $2.0d0))
+        (t
+         (let ((n (integer-length x)))
+           (cond ((< n sb-vm:double-float-digits)
+                  (log (coerce x 'double-float) $2.0d0))
+                 (t
+                  (let ((f (ldb (byte sb-vm:double-float-digits
+                                      (- n sb-vm:double-float-digits))
+                                x)))
+                    (+ n (log (scale-float (coerce f 'double-float)
+                                           (- sb-vm:double-float-digits))
+                              $2.0d0)))))))))
 
 (defun log (number &optional (base nil base-p))
   "Return the logarithm of NUMBER in the base BASE, which defaults to e."
@@ -364,7 +377,8 @@
              $0.0f0))
         ((and (typep number '(integer (0) *))
               (typep base '(integer (0) *)))
-         (coerce (/ (log2 number) (log2 base)) 'single-float))
+         (coerce (/ (truly-the double-float (log2 number))
+                    (truly-the double-float (log2 base))) 'single-float))
         ((and (typep number 'integer) (typep base 'double-float))
          ;; No single float intermediate result
          (/ (log2 number) (log base $2.0d0)))
@@ -375,18 +389,20 @@
       (number-dispatch ((number number))
         (((foreach fixnum bignum))
          (if (minusp number)
-             (complex (log (- number)) (coerce sb-xc:pi 'single-float))
-             (coerce (/ (log2 number) (log (exp $1.0d0) $2.0d0)) 'single-float)))
+             (complex (log (- number)) (coerce pi 'single-float))
+             (coerce (/ (truly-the double-float (log2 number))
+                        (log (exp $1.0d0) $2.0d0)) 'single-float)))
         ((ratio)
          (if (minusp number)
-             (complex (log (- number)) (coerce sb-xc:pi 'single-float))
+             (complex (log (- number)) (coerce pi 'single-float))
              (let ((numerator (numerator number))
                    (denominator (denominator number)))
                (if (= (integer-length numerator)
                       (integer-length denominator))
                    (coerce (%log1p (coerce (- number 1) 'double-float))
                            'single-float)
-                   (coerce (/ (- (log2 numerator) (log2 denominator))
+                   (coerce (/ (- (truly-the double-float (log2 numerator))
+                                 (truly-the double-float (log2 denominator)))
                               (log (exp $1.0d0) $2.0d0))
                            'single-float)))))
         (((foreach single-float double-float))
@@ -394,7 +410,7 @@
          ;; Since this doesn't seem to be an implementation issue
          ;; I (pw) take the Kahan result.
          (if (= (float-sign-bit number) 1) ; MINUSP
-             (complex (log (- number)) (coerce sb-xc:pi '(dispatch-type number)))
+             (complex (log (- number)) (coerce pi '(dispatch-type number)))
              (coerce (%log (coerce number 'double-float))
                      '(dispatch-type number))))
         ((complex)
@@ -449,15 +465,15 @@
   (number-dispatch ((number number))
     ((rational)
      (if (minusp number)
-         (coerce sb-xc:pi 'single-float)
+         (coerce pi 'single-float)
          $0.0f0))
     ((single-float)
      (if (minusp (float-sign number))
-         (coerce sb-xc:pi 'single-float)
+         (coerce pi 'single-float)
          $0.0f0))
     ((double-float)
      (if (minusp (float-sign number))
-         (coerce sb-xc:pi 'double-float)
+         (coerce pi 'double-float)
          $0.0d0))
     (handle-complex
      (atan (imagpart number) (realpart number)))))
@@ -548,8 +564,8 @@
                    (if (zerop y)
                        (if (= (float-sign-bit x) 0) ; PLUSP
                            y
-                           (float-sign y sb-xc:pi))
-                       (float-sign y (sb-xc:/ sb-xc:pi 2)))
+                           (float-sign y pi))
+                       (float-sign y (sb-xc:/ pi 2)))
                    (%atan2 y x))))
         (number-dispatch ((y real) (x real))
           ((double-float
@@ -750,8 +766,7 @@
         ((zerop x)
          ;; The answer is negative infinity, but we are supposed to
           ;; signal divide-by-zero, so do the actual division
-         (/ $-1.0d0 x)
-         )
+         (/ $-1.0d0 x))
         (t
           (logb-finite x))))
 
@@ -910,9 +925,9 @@
   (declare (muffle-conditions compiler-note))
   (declare (type (or rational complex) z))
   (let* (;; constants
-         (theta (sb-xc:/ (sb-xc:sqrt sb-xc:most-positive-double-float) $4.0d0))
-         (rho (sb-xc:/ $4.0d0 (sb-xc:sqrt sb-xc:most-positive-double-float)))
-         (half-pi (sb-xc:/ sb-xc:pi $2.0d0))
+         (theta (sb-xc:/ (sb-xc:sqrt most-positive-double-float) $4.0d0))
+         (rho (sb-xc:/ $4.0d0 (sb-xc:sqrt most-positive-double-float)))
+         (half-pi (sb-xc:/ pi $2.0d0))
          (rp (float (realpart z) $1.0d0))
          (beta (float-sign rp $1.0d0))
          (x (* beta rp))
