@@ -158,11 +158,11 @@ distinct from the global value. Can also be SETF."
 #-compare-and-swap-vops
 (progn
 (defun cas-symbol-%info (symbol old new)
-  (%primitive sb-vm::set-slot symbol new
+  (%primitive sb-c:set-slot symbol new
               '(setf symbol-%info) sb-vm:symbol-info-slot sb-vm:other-pointer-lowtag)
   old)
 (defun sb-vm::cas-symbol-fdefn (symbol old new)
-  (%primitive sb-vm::set-slot symbol new
+  (%primitive sb-c:set-slot symbol new
               '(setf symbol-fdefn) sb-vm:symbol-fdefn-slot sb-vm:other-pointer-lowtag)
   old))
 
@@ -186,7 +186,8 @@ distinct from the global value. Can also be SETF."
 ;; The function may choose to abort the update by returning NIL.
 (defun update-symbol-info (symbol update-fn)
   (declare (symbol symbol)
-           (type (function (t) t) update-fn))
+           (type (function (t) t) update-fn)
+           (dynamic-extent update-fn))
   (prog ((info-holder (symbol-%info symbol))
          (current-info)) ; a PACKED-INFO or NIL
    outer-restart
@@ -239,6 +240,7 @@ distinct from the global value. Can also be SETF."
 (defun %ensure-plist-holder (symbol info)
   ;; Invoked only when SYMBOL is known to be a symbol.
   (declare (optimize (safety 0)))
+  (declare (sb-c::tlab :system)) ; holder must be in the heap
   (if (consp info) ; it's fine to call this with a cell already installed
       info ; all done
       (let (newcell)
@@ -376,15 +378,40 @@ distinct from the global value. Can also be SETF."
 
 (defun %make-symbol (kind name)
   (declare (ignorable kind) (type simple-string name))
+  ;; This constructor assumes that you should never create a symbol in an arena.
+  ;; That being the case, you also never want to have its name in an arena,
+  ;; or else it would make a heap->arena reference. Theoretically gensyms could
+  ;; go in an arena, but that's such an obscure case. This has worked well so far
+  ;; with the system TLAB assumption that I'm leaving that as-is, and mitigating the
+  ;; problem with arena-allocated strings, rather than guessing whether you really
+  ;; did want an arena-allocated symbol.
   (declare (sb-c::tlab :system))
-  ;; Avoid writing to the string header if it's already flagged as readonly, or off-heap.
-  (when (and (not (logtest (ash sb-vm:+vector-shareable+ 8) (get-header-data name)))
-             (dynamic-space-obj-p name))
-    (logior-array-flags name sb-vm:+vector-shareable+)) ; Set "logically read-only" bit
-  (let ((symbol
+  (binding*
+      ((name
+        ;; We clearly have permission to copy: "It is implementation-dependent whether
+        ;; the string that becomes the new-symbol's name is the given name"
+        ;; but may or may not have leeway to change the element-type.
+        ;; I don't feel like writing a different variant of POSSIBLY-FROB-TO-HEAP just
+        ;; to satisfy pedants. However, only users of arenas would detect
+        ;; a change of element-type, if they care at all, which they don't.
+        (if (or (dynamic-space-obj-p name) (read-only-space-obj-p name))
+            name ; use as-is
+            (possibly-base-stringize-to-heap name)))
+       (()
+        (when (and (not (logtest (ash sb-vm:+vector-shareable+ 8) (get-header-data name)))
+                   ;; Readonly space is physically unwritable. Don't touch it.
+                   (not (read-only-space-obj-p name)))
+          (logior-array-flags name sb-vm:+vector-shareable+))) ; Set "logically read-only" bit
+       (symbol
          (truly-the symbol
-          #+immobile-symbols (sb-vm::make-immobile-symbol name)
+          ;; If no immobile-space, easy: all symbols go in dynamic-space
           #-immobile-space (sb-vm::%%make-symbol name)
+          ;; If #+immobile-symbols, then uninterned symbols go in dynamic space, but
+          ;; interned symbols go in immobile space. Good luck IMPORTing an uninterned symbol-
+          ;; it'll work at least superficially, but if used as a code constant, the symbol's
+          ;; address may violate the assumption that it's an imm32 operand.
+          #+immobile-symbols
+          (if (eql kind 0) (sb-vm::%%make-symbol name) (sb-vm::make-immobile-symbol name))
           #+(and immobile-space (not immobile-symbols))
           (if (or (eql kind 1) ; keyword
                   (and (eql kind 2) ; random interned symbol
@@ -393,7 +420,9 @@ distinct from the global value. Can also be SETF."
                        (char= (char name (1- (length name))) #\*)))
               (sb-vm::make-immobile-symbol name)
               (sb-vm::%%make-symbol name)))))
-    (%set-symbol-package symbol nil)
+    ;; Compact-symbol (which is equivalent to #+64-bit) has the package already NIL
+    ;; because the PACKAGE-ID-BITS field defaults to 0.
+    #-compact-symbol (%set-symbol-package symbol nil)
     symbol))
 
 (defun get (symbol indicator &optional (default nil))
@@ -415,6 +444,7 @@ distinct from the global value. Can also be SETF."
 (defun %put (symbol indicator value)
   "The VALUE is added as a property of SYMBOL under the specified INDICATOR.
   Returns VALUE."
+  (declare (sb-c::tlab :system))
   (do ((pl (symbol-plist symbol) (cddr pl)))
       ((endp pl)
        (setf (symbol-plist symbol)
@@ -461,6 +491,7 @@ distinct from the global value. Can also be SETF."
           ((eq (car plist) indicator)
            (return (cadr plist))))))
 
+;;; Note: this will cons in an arena if you're using one.
 (defun %putf (place property new-value)
   (declare (type list place))
   (do ((plist place (cddr plist)))
@@ -491,12 +522,14 @@ distinct from the global value. Can also be SETF."
   nor fbound and has no properties, else it has a copy of SYMBOL's
   function, value and property list."
   (declare (type symbol symbol))
+  (declare (sb-c::tlab :system)) ; heap-cons the property list if copying it
   (setq new-symbol (make-symbol (symbol-name symbol)))
   (when copy-props
     (%set-symbol-value new-symbol
                        (%primitive sb-c:fast-symbol-value symbol))
-    (setf (symbol-plist new-symbol)
-          (copy-list (symbol-plist symbol)))
+    (locally (declare (optimize speed)) ; will inline COPY-LIST
+      (setf (symbol-plist new-symbol)
+            (copy-list (symbol-plist symbol))))
     (when (fboundp symbol)
       (setf (symbol-function new-symbol) (symbol-function symbol))))
   new-symbol)
@@ -555,6 +588,13 @@ distinct from the global value. Can also be SETF."
             (intern (%symbol-nameify prefix (incf *gentemp-counter*)) package)
           (unless accessibility (return sym))))))
 
+(macrolet ((signal-type-error (action-description)
+             `(let ((spec (type-specifier type)))
+                (error 'simple-type-error
+                       :format-control "~@<Cannot ~@? to ~S, not of type ~S.~:@>"
+                       :format-arguments (list ,action-description symbol new-value spec)
+                       :datum new-value
+                       :expected-type spec))))
 ;;; This function is to be called just before a change which would affect the
 ;;; symbol value. We don't absolutely have to call this function before such
 ;;; changes, since such changes to constants are given as undefined behavior,
@@ -567,6 +607,10 @@ distinct from the global value. Can also be SETF."
 ;;;   foo => 13, (constantp 'foo) => t
 ;;;
 ;;; ...in which case you frankly deserve to lose.
+;;;
+;;; If this function returns normally and was called from PROGV in unsafe code,
+;;; then we'll remember that it's OK not to call it again.
+;;; However, in safe code, PROGV still performs its type checking.
 (defun about-to-modify-symbol-value (symbol action &optional (new-value nil valuep) bind)
   (declare (symbol symbol))
   (declare (explicit-check))
@@ -578,7 +622,7 @@ distinct from the global value. Can also be SETF."
              (defconstant "define ~S as a constant")
              (makunbound "make ~S unbound"))))
     (let ((kind (info :variable :kind symbol)))
-      (multiple-value-bind (what continue)
+      (multiple-value-bind (complaint continuable)
           (cond ((eq kind :constant)
                  (cond ((eq symbol t)
                         (values "Veritas aeterna. (can't ~@?)" nil))
@@ -599,21 +643,46 @@ distinct from the global value. Can also be SETF."
                  (with-single-package-locked-error (:symbol symbol "unbinding the symbol ~A")
                    (when (eq (info :variable :always-bound symbol) :always-bound)
                      (values "Can't ~@?" nil)))))
-        (when what
-          (if continue
-              (cerror "Modify the constant." what (describe-action) symbol)
-              (error what (describe-action) symbol)))
-        (when valuep
-          (multiple-value-bind (type declaredp) (info :variable :type symbol)
-            ;; If globaldb returned the default of *UNIVERSAL-TYPE*,
-            ;; don't bother with a type test.
-            (when (and declaredp (not (%%typep new-value type 'functionp)))
-              (let ((spec (type-specifier type)))
-                (error 'simple-type-error
-                       :format-control "~@<Cannot ~@? to ~S, not of type ~S.~:@>"
-                       :format-arguments (list (describe-action) symbol new-value spec)
-                       :datum new-value
-                       :expected-type spec)))))))
+        (cond ((not complaint)
+               ;; the optimize bit says whether PROGV can be optimized, and not other actions
+               (when (eq action 'progv)
+                 (let ((package (symbol-package symbol)))
+                   (if (or (not package) (not (package-locked-p package)))
+                       (logior-header-bits symbol sb-vm::+symbol-fast-bindable+)))))
+              (continuable
+               (cerror "Modify the constant." complaint (describe-action) symbol))
+              (t
+               (error complaint (describe-action) symbol)))))
+    (when (and valuep (neq action 'progv))
+      (multiple-value-bind (type declaredp) (info :variable :type symbol)
+        ;; If globaldb returned the default of *UNIVERSAL-TYPE*,
+        ;; don't bother with a type test.
+        (when (and declaredp (not (%%typep new-value type 'functionp)))
+          (signal-type-error (describe-action)))))
     nil))
 
+;;; Despite the naming symmetry, these functions are not exactly symmetrical in
+;;; how they are used by the translation of PROGV
+;;; In safe code, the fast-bindable bit is checked inside the assertion,
+;;; because in all cases, we potentially perform a type-check which is too much
+;;; to inline into PROGV.
+;;; In unsafe code, if the symbol's fast-bindable bit is on, then we do NOT call
+;;; the assertion function. When called, its role is to assert that the symbol is
+;;; bindable and then set the bit saying never to call it again unless
+;;; the bit gets unset.
+(defun assert-dynbindable-safe (symbol new-value)
+  (declare (symbol symbol))
+  (unless (test-header-data-bit symbol sb-vm::+symbol-fast-bindable+)
+    (about-to-modify-symbol-value symbol 'progv nil t))
+  ;; Perform the type check here, not in ABOUT-TO-MODIFY-SYMBOL-VALUE, so that that
+  ;; function does not have to be informed when NOT to peform a check (i.e. in usafe code).
+  ;; Specifically, it can always bypass a type-check when the action is progv.
+  (multiple-value-bind (type declaredp) (info :variable :type symbol)
+    (when (and declaredp (not (%%typep new-value type 'functionp)))
+      (signal-type-error "bind ~S"))))
+(defun assert-dynbindable-unsafe (symbol)
+  (declare (symbol symbol))
+  (about-to-modify-symbol-value symbol 'progv nil t))
+) ; end MACROLET
+
 #+sb-thread (defun symbol-tls-index (x) (symbol-tls-index x)) ; necessary stub
