@@ -87,7 +87,7 @@
   (:translate make-fdefn)
   (:generator 37
     (with-fixed-allocation (result lr fdefn-widetag fdefn-size)
-      (load-inline-constant temp '(:fixup undefined-tramp :assembly-routine))
+      (load-asm-routine temp 'undefined-tramp)
       (storew name result fdefn-name-slot other-pointer-lowtag)
       (storew null-tn result fdefn-fun-slot other-pointer-lowtag)
       (storew temp result fdefn-raw-addr-slot other-pointer-lowtag))))
@@ -109,10 +109,16 @@
         (inst adr lr label (ash simple-fun-insts-offset word-shift))
         (storew-pair temp 0 lr closure-fun-slot tmp-tn)))))
 
+(define-vop (reference-closure)
+  (:info label)
+  (:results (result :scs (descriptor-reg)))
+  (:generator 1
+    (inst adr result label fun-pointer-lowtag)))
+
 ;;; The compiler likes to be able to directly make value cells.
 ;;;
 (define-vop (make-value-cell)
-  (:args (value :to :save :scs (descriptor-reg any-reg)))
+  (:args (value :to :save :scs (descriptor-reg any-reg zero)))
   (:temporary (:scs (non-descriptor-reg) :offset lr-offset) lr)
   (:info stack-allocate-p)
   (:results (result :scs (descriptor-reg)))
@@ -134,7 +140,7 @@
   (:args)
   (:results (result :scs (any-reg)))
   (:generator 1
-    (load-inline-constant result '(:fixup funcallable-instance-tramp :assembly-routine))))
+    (load-asm-routine result 'funcallable-instance-tramp)))
 
 (define-vop (fixed-alloc)
   (:args)
@@ -148,7 +154,7 @@
                             :stack-allocate-p stack-allocate-p))))
 
 (define-vop (var-alloc)
-  (:args (extra :scs (any-reg)))
+  (:args (extra :scs (any-reg) :target bytes))
   (:arg-types positive-fixnum)
   (:info name words type lowtag stack-allocate-p)
   (:ignore name stack-allocate-p)
@@ -160,7 +166,9 @@
     ;; Build the object header, assuming that the header was in WORDS
     ;; but should not be in the header
     (inst lsl bytes extra (- word-shift n-fixnum-tag-bits))
-    (inst add bytes bytes (add-sub-immediate (* (1- words) n-word-bytes)))
+    (let ((words (add-sub-immediate (* (1- words) n-word-bytes))))
+      (unless (eql words 0)
+        (inst add bytes bytes words)))
     (inst lsl header bytes (- (length-field-shift type) word-shift))
     (inst add header header type)
     ;; Add the object header to the allocation size and round up to
@@ -194,9 +202,8 @@
      (move c-arg1 size-class)
      (move c-arg2 nwords)
      (move c-arg3 header)
-     (load-inline-constant lr '(:fixup "call_into_c" :foreign))
-     (load-inline-constant cfunc '(:fixup "alloc_immobile_fixedobj" :foreign))
-     (inst blr lr)
+     (load-foreign-symbol cfunc "alloc_immobile_fixedobj")
+     (invoke-foreign-routine "call_into_c" lr)
      (when cur-nfp
        (load-stack-tn cur-nfp nfp-save))
      (move result nl0))))
