@@ -934,13 +934,18 @@ also processed as top level forms."
     (cond ((some #'leaf-dynamic-extent funs)
            (ctran-starts-block next)
            (let* ((enclose (ctran-use enclose-ctran))
+                  (dynamic-extent (make-dynamic-extent))
                   (cleanup (make-cleanup :kind :dynamic-extent
-                                         :mess-up enclose))
-                  (cleanup-ctran (make-ctran)))
-             (setf (enclose-cleanup enclose) cleanup)
+                                         :mess-up dynamic-extent))
+                  (dynamic-extent-ctran (make-ctran)))
+             (setf (enclose-dynamic-extent enclose) dynamic-extent)
+             (setf (dynamic-extent-cleanup dynamic-extent) cleanup)
+             (link-node-to-previous-ctran dynamic-extent enclose-ctran)
+             (use-ctran dynamic-extent dynamic-extent-ctran)
+             (push dynamic-extent
+                   (lambda-dynamic-extents (node-home-lambda dynamic-extent)))
              (let ((*lexenv* (make-lexenv :cleanup cleanup)))
-               (ir1-convert enclose-ctran cleanup-ctran nil '(%cleanup-point))
-               (ir1-convert-progn-body cleanup-ctran next result body))))
+               (ir1-convert-progn-body dynamic-extent-ctran next result body))))
           (t
            (ir1-convert-progn-body enclose-ctran next result body)))))
 
@@ -1538,8 +1543,8 @@ values from the first VALUES-FORM making up the first argument, etc."
 ;;; to interpose a DELAY node using RESULT immediately so that the
 ;;; result continuation can assume that it is immediately used. This
 ;;; is important here because MULTIPLE-VALUE-PROG1 is the only special
-;;; form which receives unknown values with multiple uses, some (in
-;;; this case one) of which are not immediate.
+;;; form that produces code where lvar substitution is potentially
+;;; incorrect.
 (def-ir1-translator multiple-value-prog1
     ((values-form &rest forms) start next result)
   "MULTIPLE-VALUE-PROG1 values-form form*
