@@ -472,7 +472,6 @@ necessary, since type inference may take arbitrarily long to converge.")
                ;; information didn't lead to any new optimizations.
                ;; Don't run constraint-propagate again.
                (return)))
-        (eliminate-dead-code component)
         (dfo-as-needed component)
         (when constraint-propagate
           (maybe-mumble "Constraint ")
@@ -577,7 +576,7 @@ necessary, since type inference may take arbitrarily long to converge.")
   (report-code-deletion)
 
   (when (or (ir2-component-values-receivers (component-info component))
-            (component-dx-lvars component))
+            (ir2-component-stack-allocates-p (component-info component)))
     (maybe-mumble "Stack ")
     (stack-analyze component)
     ;; Assign BLOCK-NUMBER for any cleanup blocks introduced by
@@ -713,10 +712,11 @@ necessary, since type inference may take arbitrarily long to converge.")
 
     (ir1-phases component)
 
-    ;; This should happen at some point before ENVIRONMENT-ANALYZE,
-    ;; and after RECORD-COMPONENT-XREFS.  Beyond that, I haven't
-    ;; really thought things through.  -- AJB, 2014-Jun-08
-    (eliminate-dead-code component)
+    ;; KLUDGE: We should instead set COMPONENT-REOPTIMIZE to T
+    ;; whenever a REF gets deleted so that DFO-AS-NEEDED kicks in only
+    ;; when needed, and we don't need this call to do a final
+    ;; unreachable entry point scan.
+    (find-dfo component)
 
     (dfo-as-needed component)
     (maybe-mumble "Dom ")
@@ -949,8 +949,7 @@ necessary, since type inference may take arbitrarily long to converge.")
 ;; a subtype of not-so-aptly-named INPUT-ERROR-IN-COMPILE-FILE.
 (defun %do-forms-from-info (function info condition-name)
   (declare (function function))
-  ; FIXME: why "could not stack-allocate" in a few call sites?
-  ; (declare (dynamic-extent function))
+  (declare (dynamic-extent function))
   (let* ((file-info (source-info-file-info info))
          (stream (get-source-stream info))
          (pos (file-position stream))
@@ -1144,24 +1143,6 @@ necessary, since type inference may take arbitrarily long to converge.")
           (intersection '(:load-toplevel load) situations)
           (intersection '(:execute eval) situations)))
 
-
-;;; utilities for extracting COMPONENTs of FUNCTIONALs
-(defun functional-components (f)
-  (declare (type functional f))
-  (etypecase f
-    (clambda (list (lambda-component f)))
-    (optional-dispatch (let ((result nil))
-                         (flet ((maybe-frob (maybe-clambda)
-                                  (when (and maybe-clambda
-                                             (promise-ready-p maybe-clambda))
-                                    (pushnew (lambda-component
-                                              (force maybe-clambda))
-                                             result))))
-                           (map nil #'maybe-frob (optional-dispatch-entry-points f))
-                           (maybe-frob (optional-dispatch-more-entry f))
-                           (maybe-frob (optional-dispatch-main-entry f)))
-                         result))))
-
 ;;; Print some noise about FORM if *COMPILE-PRINT* is true.
 (defun note-top-level-form (form)
   (when *compile-print*
@@ -1583,8 +1564,6 @@ necessary, since type inference may take arbitrarily long to converge.")
           (compile-load-time-value-lambda lambdas)
           (compile-toplevel-lambdas lambdas top-level-closure)))
 
-    (dolist (component components)
-      (clear-ir1-info component))
     (clear-ir1-namespace))
   (values))
 
