@@ -205,39 +205,43 @@
   ;; (Ideally we'd only do 1 pointer bump, but that's a separate issue)
   (inst test :byte rax-tn (ea (- static-space-start gc-safepoint-trap-offset))))
 
+(macrolet ((pa-bits-ea ()
+             #+sb-thread `(thread-slot-ea
+                           thread-pseudo-atomic-bits-slot
+                           #+gs-seg ,@(if thread (list thread)))
+             #-sb-thread `(static-symbol-value-ea '*pseudo-atomic-bits*))
+           (nonzero-bits ()
+             ;; reg-mem move is allegedly faster than imm-mem according to
+             ;; someone at some point. Whether that's true or not, it is what it is.
+             ;; THREAD-TN is a better choice than RBP-TN since it's constant.
+             #+(and sb-thread (not gs-seg)) 'thread-tn
+             #-(and sb-thread (not gs-seg)) 'rbp-tn))
+  (defun emit-begin-pseudo-atomic ()
+    #-sb-safepoint (inst mov (pa-bits-ea) (nonzero-bits)))
+  (defun emit-end-pseudo-atomic ()
+    #+sb-safepoint (emit-safepoint)
+    #-sb-safepoint
+    (assemble ()
+      (inst xor (pa-bits-ea) (nonzero-bits))
+      (inst jmp :z OUT)
+      ;; if PAI was set, interrupts were disabled at the same time
+      ;; using the process signal mask.
+      #+int1-breakpoints (inst icebp)
+      #-int1-breakpoints (inst break pending-interrupt-trap)
+      OUT)))
+
 ;;; This macro is purposely unhygienic with respect to THREAD-TN,
 ;;; which is either a global symbol macro, or a LET-bound variable,
 ;;; depending on #+gs-seg.
-(defmacro pseudo-atomic ((&key ((:thread-tn thread)) elide-if) &body forms)
+(defmacro pseudo-atomic ((&key ((:thread-tn thread)) elide-if (default-exit t))
+                         &body forms)
   (declare (ignorable thread))
-  #+sb-safepoint
-  `(progn ,@forms (unless ,elide-if (emit-safepoint)))
-  #-sb-safepoint
-  (with-unique-names (label pa-bits-ea)
-   (let ((true
-          ;; TRUE is anything nonzero. Moving a register to memory is
-          ;; allegedly faster than reading an imm8 operand. I don't know,
-          ;; but I'm not going to debate it. However THREAD-TN is a better
-          ;; choice than RBP-TN since it's never written to.
-          #+(and sb-thread (not gs-seg)) 'thread-tn
-          #-(and sb-thread (not gs-seg)) 'rbp-tn))
-    `(let ((,label (gen-label))
-           (,pa-bits-ea
-            #+sb-thread (thread-slot-ea
-                         thread-pseudo-atomic-bits-slot
-                         #+gs-seg ,@(if thread (list thread)))
-            #-sb-thread (static-symbol-value-ea '*pseudo-atomic-bits*)))
-       (unless ,elide-if
-         (inst mov ,pa-bits-ea ,true))
-       ,@forms
-       (unless ,elide-if
-         (inst xor ,pa-bits-ea ,true)
-         (inst jmp :z ,label)
-         ;; if PAI was set, interrupts were disabled at the same time
-         ;; using the process signal mask.
-         #+int1-breakpoints (inst icebp)
-         #-int1-breakpoints (inst break pending-interrupt-trap)
-         (emit-label ,label))))))
+  `(macrolet ((exit-pseudo-atomic () '(emit-end-pseudo-atomic)))
+     (unless ,elide-if
+       (emit-begin-pseudo-atomic))
+     (assemble () ,@forms)
+     (when (and ,default-exit (not ,elide-if))
+       (exit-pseudo-atomic))))
 
 ;;;; indexed references
 
@@ -267,7 +271,7 @@
               (index :scs (,@(when (member translate '(%instance-cas %raw-instance-cas/word))
                                '(immediate))
                            any-reg signed-reg unsigned-reg) :to :eval)
-              (old-value :scs ,scs #|:target rax|#)
+              (old-value :scs (,@scs immediate) #|:target rax|#)
               (new-value :scs ,scs))
        (:vop-var vop)
        (:arg-types ,type tagged-num ,el-type ,el-type)
@@ -288,12 +292,12 @@
            ,@(ecase name
                (%compare-and-swap-svref
                 ;; store barrier needs the EA of the affected element
-                '((emit-gc-store-barrier object ea rax (vop-nth-arg 3 vop) new-value)))
+                '((emit-gengc-barrier object ea rax (vop-nth-arg 3 vop) new-value)))
                (%instance-cas
                 ;; store barrier affects only the object's base address
-                '((emit-gc-store-barrier object nil rax (vop-nth-arg 3 vop) new-value)))
+                '((emit-gengc-barrier object nil rax (vop-nth-arg 3 vop) new-value)))
                ((%raw-instance-cas/word %raw-instance-cas/signed-word)))
-           (move rax old-value)
+           (move-immediate rax (encode-value-if-immediate old-value ,(and (memq 'any-reg scs) t)))
            (inst cmpxchg :lock ea new-value)
            (move value rax))))))
 
@@ -425,6 +429,11 @@
                            object)
                        (ea (- (* ,offset n-word-bytes) ,lowtag)
                            object index (index-scale n-word-bytes index)))))
-           ,@(when (member name '(instance-index-set %closure-index-set))
-               '((emit-gc-store-barrier object nil val-temp (vop-nth-arg 2 vop) value)))
-           (gen-cell-set ea value val-temp)))))
+           ,@(when (member name '(instance-index-set %closure-index-set %weakvec-set))
+               '((emit-gengc-barrier object nil val-temp (vop-nth-arg 2 vop) value)))
+           (emit-store ea value val-temp)))))
+
+(defmacro pc-size (vop)
+  `(if (sb-c::code-immobile-p ,vop)
+       :dword
+       :qword))
